from prometheus_client import Counter
from prometheus_client import Histogram

def init_metrics(registry):
  return {
      "grpc_client_started_counter": Counter(
          "grpc_client_started_total",
          "Total number of RPCs started on the client",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),

      "grpc_client_handled_counter": Counter(
          "grpc_client_handled_total",
          "Total number of RPCs completed on the client, "
          "regardless of success or failure.",
          ["grpc_type", "grpc_service", "grpc_method", "grpc_code"],
          registry=registry
      ),

      "grpc_client_stream_msg_received": Counter(
          "grpc_client_msg_received_total",
          "Total number of RPC stream messages received by the client.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),

      "grpc_client_stream_msg_sent": Counter(
          "grpc_client_msg_sent_total",
          "Total number of gRPC stream messages sent by the client.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),

      "grpc_client_handled_histogram": Histogram(
          "grpc_client_handling_seconds",
          "Histogram of response latency (seconds) of the gRPC until" \
            "it is finished by the application.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),

      "grpc_client_stream_recv_histogram": Histogram(
          "grpc_client_msg_recv_handling_seconds",
          "Histogram of response latency (seconds) of the gRPC single message receive.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),

      "grpc_client_stream_send_histogram": Histogram(
          "grpc_client_msg_send_handling_seconds",
          "Histogram of response latency (seconds) of the gRPC single message send.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),

      # Legacy metrics for backwards compatibility

      "legacy_grpc_client_completed_counter": Counter(
          "grpc_client_completed",
          "Total number of RPCs completed on the client, "
          "regardless of success or failure.",
          ["grpc_type", "grpc_service", "grpc_method", "code"],
          registry=registry
      ),

      "legacy_grpc_client_completed_latency_seconds_histogram": Histogram(
          "grpc_client_completed_latency_seconds",
          "Histogram of rpc response latency (in seconds) for completed rpcs.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),
  }



UNARY = "UNARY"
SERVER_STREAMING = "SERVER_STREAMING"
CLIENT_STREAMING = "CLIENT_STREAMING"
BIDI_STREAMING = "BIDI_STREAMING"
UNKNOWN = "UNKNOWN"


def wrap_iterator_inc_counter(iterator, counter, grpc_type, grpc_service_name, grpc_method_name):
  """Wraps an iterator and collect metrics."""

  for item in iterator:
    counter.labels(
      grpc_type=grpc_type,
      grpc_service=grpc_service_name,
      grpc_method=grpc_method_name).inc()
    yield item


def get_method_type(request_streaming, response_streaming):
  """
  Infers the method type from if the request or the response is streaming.

  # The Method type is coming from:
  # https://grpc.io/grpc-java/javadoc/io/grpc/MethodDescriptor.MethodType.html
  """
  if request_streaming and response_streaming:
    return BIDI_STREAMING
  elif request_streaming and not response_streaming:
    return CLIENT_STREAMING
  elif not request_streaming and response_streaming:
    return SERVER_STREAMING
  return UNARY


def split_method_call(handler_call_details):
  """
  Infers the grpc service and method name from the handler_call_details.
  """

  # e.g. /package.ServiceName/MethodName
  parts = handler_call_details.method.split("/")
  if len(parts) < 3:
    return "", "", False

  grpc_service_name, grpc_method_name = parts[1:3]
  return grpc_service_name, grpc_method_name, True



"""Interceptor a client call with prometheus"""

from timeit import default_timer

import grpc
from prometheus_client.registry import REGISTRY

from py_grpc_prometheus import grpc_utils
from py_grpc_prometheus.client_metrics import init_metrics

class PromClientInterceptor(grpc.UnaryUnaryClientInterceptor,
                            grpc.UnaryStreamClientInterceptor,
                            grpc.StreamUnaryClientInterceptor,
                            grpc.StreamStreamClientInterceptor):
  """
  Intercept gRPC client requests.
  """

  def __init__(
            self,
            enable_client_handling_time_histogram=False,
            enable_client_stream_receive_time_histogram=False,
            enable_client_stream_send_time_histogram=False,
            legacy=False,
            registry=REGISTRY
  ):
    self._enable_client_handling_time_histogram = enable_client_handling_time_histogram
    self._enable_client_stream_receive_time_histogram = enable_client_stream_receive_time_histogram
    self._enable_client_stream_send_time_histogram = enable_client_stream_send_time_histogram
    self._legacy = legacy
    self._metrics = init_metrics(registry)

  def intercept_unary_unary(self, continuation, client_call_details, request):
    grpc_service_name, grpc_method_name, _ = grpc_utils.split_method_call(client_call_details)
    grpc_type = grpc_utils.UNARY


    self._metrics["grpc_client_started_counter"].labels(
        grpc_type=grpc_type,
        grpc_service=grpc_service_name,
        grpc_method=grpc_method_name).inc()

    start = default_timer()
    handler = continuation(client_call_details, request)
    if self._legacy:
      self._metrics["legacy_grpc_client_completed_latency_seconds_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))
    elif self._enable_client_handling_time_histogram:
      self._metrics["grpc_client_handled_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    if self._legacy:
      self._metrics["legacy_grpc_client_completed_counter"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name,
          code=handler.code().name).inc()
    else:
      self._metrics["grpc_client_handled_counter"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name,
          grpc_code=handler.code().name).inc()

    return handler

  def intercept_unary_stream(self, continuation, client_call_details, request):
    grpc_service_name, grpc_method_name, _ = grpc_utils.split_method_call(client_call_details)
    grpc_type = grpc_utils.SERVER_STREAMING

    self._metrics["grpc_client_started_counter"].labels(
        grpc_type=grpc_type,
        grpc_service=grpc_service_name,
        grpc_method=grpc_method_name).inc()

    start = default_timer()
    handler = continuation(client_call_details, request)
    if self._legacy:
      self._metrics["legacy_grpc_client_completed_latency_seconds_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    elif self._enable_client_handling_time_histogram:
      self._metrics["grpc_client_handled_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    handler = grpc_utils.wrap_iterator_inc_counter(
        handler,
        self._metrics["grpc_client_stream_msg_received"],
        grpc_type,
        grpc_service_name,
        grpc_method_name)

    if self._enable_client_stream_receive_time_histogram and not self._legacy:
      self._metrics["grpc_client_stream_recv_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    return handler

  def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
    grpc_service_name, grpc_method_name, _ = grpc_utils.split_method_call(client_call_details)
    grpc_type = grpc_utils.CLIENT_STREAMING

    iterator_metric = self._metrics["grpc_client_stream_msg_sent"]

    request_iterator = grpc_utils.wrap_iterator_inc_counter(
        request_iterator,
        iterator_metric,
        grpc_type,
        grpc_service_name,
        grpc_method_name)

    start = default_timer()
    handler = continuation(client_call_details, request_iterator)

    if self._legacy:
      self._metrics["grpc_client_started_counter"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).inc()
      self._metrics["legacy_grpc_client_completed_latency_seconds_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))
    else:
      self._metrics["grpc_client_started_counter"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).inc()
      if self._enable_client_handling_time_histogram:
        self._metrics["grpc_client_handled_histogram"].labels(
            grpc_type=grpc_type,
            grpc_service=grpc_service_name,
            grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    if self._enable_client_stream_send_time_histogram and not self._legacy:
      self._metrics["grpc_client_stream_send_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    return handler

  def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
    grpc_service_name, grpc_method_name, _ = grpc_utils.split_method_call(
        client_call_details)
    grpc_type = grpc_utils.BIDI_STREAMING
    start = default_timer()

    iterator_sent_metric = self._metrics["grpc_client_stream_msg_sent"]

    response_iterator = continuation(
        client_call_details,
        grpc_utils.wrap_iterator_inc_counter(
            request_iterator,
            iterator_sent_metric,
            grpc_type,
            grpc_service_name,
            grpc_method_name))

    if self._enable_client_stream_send_time_histogram and not self._legacy:
      self._metrics["grpc_client_stream_send_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    iterator_received_metric = self._metrics["grpc_client_stream_msg_received"]

    response_iterator = grpc_utils.wrap_iterator_inc_counter(
        response_iterator,
        iterator_received_metric,
        grpc_type,
        grpc_service_name,
        grpc_method_name)

    if self._enable_client_stream_receive_time_histogram and not self._legacy:
      self._metrics["grpc_client_stream_recv_histogram"].labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name).observe(max(default_timer() - start, 0))

    return response_iterator



from prometheus_client import Counter
from prometheus_client import Histogram

def init_metrics(registry):
  return {
      "grpc_server_started_counter": Counter(
          "grpc_server_started_total",
          "Total number of RPCs started on the server.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),
      "grpc_server_stream_msg_received": Counter(
          "grpc_server_msg_received_total",
          "Total number of RPC stream messages received on the server.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),
      "grpc_server_stream_msg_sent": Counter(
          "grpc_server_msg_sent_total",
          "Total number of gRPC stream messages sent by the server.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),
      "grpc_server_handled_histogram": Histogram(
          "grpc_server_handling_seconds",
          "Histogram of response latency (seconds) of gRPC that had been application-level "
          "handled by the server.",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      ),
      "legacy_grpc_server_handled_latency_seconds": Histogram(
          "grpc_server_handled_latency_seconds",
          "Histogram of response latency (seconds) of gRPC that had been "
          "application-level handled by the server",
          ["grpc_type", "grpc_service", "grpc_method"],
          registry=registry
      )
  }


# Legacy metrics for backward compatibility
def get_grpc_server_handled_counter(is_legacy, registry):
  if is_legacy:
    return Counter(
        "grpc_server_handled_total",
        "Total number of RPCs completed on the server, regardless of success or failure.",
        ["grpc_type", "grpc_service", "grpc_method", "code"],
        registry=registry
    )
  return Counter(
      "grpc_server_handled_total",
      "Total number of RPCs completed on the server, regardless of success or failure.",
      ["grpc_type", "grpc_service", "grpc_method", "grpc_code"],
      registry=registry
  )



#!/usr/bin/env python
from setuptools import find_packages
from setuptools import setup

with open("README.md", "r") as fh:
  long_description = fh.read()

setup(name="py_grpc_prometheus",
      version="0.7.0",
      description="Python gRPC Prometheus Interceptors",
      long_description=long_description,
      long_description_content_type="text/markdown",
      author="Lin Chen",
      author_email="linchen04@gmail.com",
      install_requires=[
          "setuptools>=39.0.1",
          "grpcio>=1.10.0",
          "prometheus_client>=0.3.0"
      ],
      url="https://github.com/lchenn/py-grpc-prometheus",
      packages=find_packages(exclude=["tests.*", "tests"]),
     )



"""Interceptor a client call with prometheus"""
import logging

from timeit import default_timer

import grpc
from prometheus_client.registry import REGISTRY

from py_grpc_prometheus import grpc_utils
from py_grpc_prometheus import server_metrics


_LOGGER = logging.getLogger(__name__)

class PromServerInterceptor(grpc.ServerInterceptor):

  def __init__(self,
               enable_handling_time_histogram=False,
               legacy=False,
               skip_exceptions=False,
               log_exceptions=True,
               registry=REGISTRY):
    self._enable_handling_time_histogram = enable_handling_time_histogram
    self._legacy = legacy
    self._grpc_server_handled_total_counter = server_metrics.get_grpc_server_handled_counter(
        self._legacy,
        registry
    )
    self._metrics = server_metrics.init_metrics(registry)
    self._skip_exceptions = skip_exceptions
    self._log_exceptions = log_exceptions

  def intercept_service(self, continuation, handler_call_details):
    """
    Intercepts the server function calls.

    This implements referred to:
    https://github.com/census-instrumentation/opencensus-python/blob/master/opencensus/
    trace/ext/grpc/server_interceptor.py
    and
    https://grpc.io/grpc/python/grpc.html#service-side-interceptor
    """

    grpc_service_name, grpc_method_name, _ = grpc_utils.split_method_call(handler_call_details)

    def metrics_wrapper(behavior, request_streaming, response_streaming):
      def new_behavior(request_or_iterator, servicer_context):
        response_or_iterator = None
        try:
          start = default_timer()
          grpc_type = grpc_utils.get_method_type(request_streaming, response_streaming)
          try:
            if request_streaming:
              request_or_iterator = grpc_utils.wrap_iterator_inc_counter(
                  request_or_iterator,
                  self._metrics["grpc_server_stream_msg_received"],
                  grpc_type,
                  grpc_service_name,
                  grpc_method_name)
            else:
              self._metrics["grpc_server_started_counter"].labels(
                  grpc_type=grpc_type,
                  grpc_service=grpc_service_name,
                  grpc_method=grpc_method_name).inc()

            # Invoke the original rpc behavior.
            response_or_iterator = behavior(request_or_iterator, servicer_context)

            if response_streaming:
              sent_metric = self._metrics["grpc_server_stream_msg_sent"]
              response_or_iterator = grpc_utils.wrap_iterator_inc_counter(
                  response_or_iterator,
                  sent_metric,
                  grpc_type,
                  grpc_service_name,
                  grpc_method_name)

            else:
              self.increase_grpc_server_handled_total_counter(grpc_type,
                                                              grpc_service_name,
                                                              grpc_method_name,
                                                              self._compute_status_code(
                                                                  servicer_context).name)
            return response_or_iterator
          except grpc.RpcError as e:
            self.increase_grpc_server_handled_total_counter(grpc_type,
                                                            grpc_service_name,
                                                            grpc_method_name,
                                                            self._compute_error_code(e).name)
            raise e

          finally:

            if not response_streaming:
              if self._legacy:
                self._metrics["legacy_grpc_server_handled_latency_seconds"].labels(
                    grpc_type=grpc_type,
                    grpc_service=grpc_service_name,
                    grpc_method=grpc_method_name) \
                    .observe(max(default_timer() - start, 0))
              elif self._enable_handling_time_histogram:
                self._metrics["grpc_server_handled_histogram"].labels(
                    grpc_type=grpc_type,
                    grpc_service=grpc_service_name,
                    grpc_method=grpc_method_name) \
                    .observe(max(default_timer() - start, 0))
        except Exception as e: # pylint: disable=broad-except
          # Allow user to skip the exceptions in order to maintain
          # the basic functionality in the server
          # The logging function in exception can be toggled with log_exceptions
          # in order to suppress the noise in logging
          if self._skip_exceptions:
            if self._log_exceptions:
              _LOGGER.error(e)
            if response_or_iterator is None:
              return response_or_iterator
            return behavior(request_or_iterator, servicer_context)
          raise e

      return new_behavior

    optional_any = self._wrap_rpc_behavior(continuation(handler_call_details), metrics_wrapper)

    return optional_any

  # pylint: disable=protected-access
  def _compute_status_code(self, servicer_context):
    if servicer_context._state.client == "cancelled":
      return grpc.StatusCode.CANCELLED

    if servicer_context._state.code is None:
      return grpc.StatusCode.OK

    return servicer_context._state.code

  def _compute_error_code(self, grpc_exception):
    if isinstance(grpc_exception, grpc.Call):
      return grpc_exception.code()

    return grpc.StatusCode.UNKNOWN

  def increase_grpc_server_handled_total_counter(
      self, grpc_type, grpc_service_name, grpc_method_name, grpc_code):
    if self._legacy:
      self._grpc_server_handled_total_counter.labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name,
          code=grpc_code).inc()
    else:
      self._grpc_server_handled_total_counter.labels(
          grpc_type=grpc_type,
          grpc_service=grpc_service_name,
          grpc_method=grpc_method_name,
          grpc_code=grpc_code).inc()

  def _wrap_rpc_behavior(self, handler, fn):
    """Returns a new rpc handler that wraps the given function"""
    if handler is None:
      return None

    if handler.request_streaming and handler.response_streaming:
      behavior_fn = handler.stream_stream
      handler_factory = grpc.stream_stream_rpc_method_handler
    elif handler.request_streaming and not handler.response_streaming:
      behavior_fn = handler.stream_unary
      handler_factory = grpc.stream_unary_rpc_method_handler
    elif not handler.request_streaming and handler.response_streaming:
      behavior_fn = handler.unary_stream
      handler_factory = grpc.unary_stream_rpc_method_handler
    else:
      behavior_fn = handler.unary_unary
      handler_factory = grpc.unary_unary_rpc_method_handler

    return handler_factory(
        fn(behavior_fn, handler.request_streaming, handler.response_streaming),
        request_deserializer=handler.request_deserializer,
        response_serializer=handler.response_serializer)
